% Demo script for the IUB Technical Report Nr 3, Herbert Jaeger:
% "Generating exponentially many periodic attractors with linearly growing Echo State
% Networks".
%
% Herbert Jaeger, August 2006
% 
% Notice: International patents pending for the Echo State Network architecture 
% and learning method, claimed by Fraunhofer IAIS
% http://www.iais.fraunhofer.de/ for most of Europe, USA, Canada,
% Australia, Japan. 

p = 5; % nr of pitch bins
d = 10; % nr of delays

% 1. training ESN

% 1.1 generate ESN
reservoirSize = 100;  
% teacherScaling must be same as inputScaling, teacherShift same as
% inputShift (except array sizes)
esn = generate_esn_8(d, reservoirSize, p,'spectralRadius',0.9,...
    'inputWeightScaling', 1.0 * ones(p,1),'bias', 0.0 * ones(p,1),...
    'teacherScaling', 0.9 * ones(p, d), 'teacherShift', 0.05 * ones(p, d),...
    'inputScaling', 0.9 * ones(p,1), 'inputShift', 0.05 * ones(p,1), ...
    'reservoirActivationFunction', 'identity_8', ...
    'dataActivationFunction', 'sigmoid01', ...
    'inverseDataActivationFunction', 'sigmoid01_inv', ...
    'methodTrain', 'wiener_hopf_8') ; 

               
% 1.2 prepare training data
sampleLength = 1000;
rawSignal = rand(sampleLength + d,1);
sampleInput = rawToUnitCoded(rawSignal, p);
sampleInput = cleanPitches(sampleInput);

% order p x d (pitch times delay) outputs in d many portions each of length
% p. Shortest delay come first
sampleOutput = zeros(sampleLength, p, d);
for i = 1:d
    sampleOutput(:,:,d-i+1) = sampleInput(i:end - d + i-1,:);
end
sampleInput = sampleInput(d + 1: end, :);

[sampleInputTrain, sampleInputTest] = split_train_test_8(sampleInput, 0.75);
[sampleOutputTrain, sampleOutputTest] = split_train_test_8(sampleOutput, 0.75);


% 1.3 train the ESN
nDiscard = 200;
trainVotes = [ones(size(sampleInputTrain,1),1) zeros(size(sampleInputTrain,1), d)];
trainNoiseLevel = 0.0001;
[esn.W_out stateMatrix teachMatrix] = ...
    train_esn_8(sampleInputTrain, sampleOutputTrain, trainVotes, esn, nDiscard, trainNoiseLevel) ; 


% 1.4 diagnostics for training
figure(1);
titlestring = 'four internal states during learning';
plot_states(stateMatrix,[1 2 3 4], 200, titlestring) ; 


figure(2);
plot(squeeze(mean(abs(esn.W_out),2)));
title('mean abs output weights');

estimatedTrainOutput = zeros(size(stateMatrix,1), p, d);
for n = 1:esn.d
    estimatedTrainOutput(:,:,n) = feval(esn.dataActivationFunction, ...
        stateMatrix * squeeze(esn.W_out(:,:,n))');
    estimatedTrainOutput(:,:,n) = ...
        estimatedTrainOutput(:,:,n) - repmat(esn.teacherShift(:,n)', size(stateMatrix,1),1);
    estimatedTrainOutput(:,:,n) = ...
        estimatedTrainOutput(:,:,n)./ repmat(esn.teacherScaling(:,n)', size(stateMatrix,1),1);
end

figure(3); clf;
outputs_plotted = [1 1; 2 3; 3 5; 4 7; 5 10]; % select some pairs from p x d
nplots = size(outputs_plotted, 1);
lengthPlot = 200;
for i = 1:nplots
    subplot(nplots, 1, i);
    plot(estimatedTrainOutput(1:lengthPlot, outputs_plotted(i,1), outputs_plotted(i,2))); hold on;
    plot(sampleOutputTrain(nDiscard+1:nDiscard+lengthPlot+1, ...
        outputs_plotted(i,1), outputs_plotted(i,2)), 'g:'); hold off;
    if i == 1
        title('trained ESN output(b) vs. teacher(g:)');
    end
end

trainNRMSE = compute_error_8(estimatedTrainOutput, sampleOutputTrain, esn); % size (p, d)
figure(4); clf;
plot(trainNRMSE);
title('training NRMSE');

%compute NMSE testing error
voteHistory = [ones(size(sampleInputTest,1),1) zeros(size(sampleInputTest,1), d)];
[testOutput, stateCollection] = test_esn_8(sampleInputTest, voteHistory, esn, nDiscard) ; 
testNRMSE = compute_error_8(testOutput, sampleOutputTest, esn); 
figure(5); clf;
plot(testNRMSE);
title('testing NRMSE');

% use delay-line ESN to continue rhythmic trigger pattern
initN = 20; periodN = 7; repeatN = 10;
triggerInputRaw = rand(initN + periodN * repeatN, 1) ;
periodPattern = rand(periodN ,1);
for r = 1:repeatN
    triggerInputRaw(initN + (r-1)*periodN + 1 : initN + r * periodN, 1) = periodPattern;
end
triggerInput = rawToUnitCoded(triggerInputRaw, esn.p);
triggerInput = cleanPitches(triggerInput);
triggerInputRaw = unitCodedToRaw(triggerInput);

zeroConfidenceThreshold = 0.3;
votingGrowthConstant = 4;
votingDecay = 0.2;
errDecay = 0.4; errGrowthConstant = 4; 
startAfterRepeatsN = 2; noiseLevel = 0.01;
netOutCompletePL = run_melody_8(esn, triggerInput, triggerInputRaw, ...
    initN, periodN, repeatN, startAfterRepeatsN, ...
    zeroConfidenceThreshold, votingGrowthConstant, votingDecay, errDecay, ...
    errGrowthConstant, noiseLevel);

