function netOutCompletePL = run_melody_8(esn, ...
    triggerInput, triggerInputRaw, ...
    initN, periodN, repeatN, ...
    startAfterRepeatsN,  zeroPredictConfidenceThreshold, votingGrowthConstant, ...
    votingDecay, errDecay, errGrowthConstant, noiseLevel)

% triggerInput is driving sequence made of random initital part of length
% initN, followed by a periodic pattern of length periodN.
% The network is run in two phases. First phase: it is driven by
% triggerInput, up to the time step initN + periodN * repeatN, that is,
% after this phase the network has seen repeatN many times the pattern. 
% In the second phase, the network does not take notice of the triggerInput
% any more (it is available to the program only for plotting it later
% against the network output). Instead, the network autonomously generates
% output, which is fed back into the input unit immediately. 
% zeroPredictConfidenceThreshold is real number between 0 and 1, indicating the
% size of prediction error where an output unit believes it has zero
% predictConfidence. A standard setting is 0.5.
% votingGrowthConstant is multiplicative constant for weighting the current
% predictConfidence vector when it is added to the voting state.
% votingDecay is exponential decay constant for predictConfidence state
% errDecay is exponential decay constant for leaky-integrated error state
% errGrowthConstant is multiplicative constant for the accumulation rate of
% the error integration
% emphasisExp is exponent (typically between 1 and 2) for putting nonlinear emphasis
% on the better-predicting network outputs
% noiseLevel is level of uniform noise added to network states during the
% entire execution, in order to check stability


totalstate = zeros(esn.reservoirSize + esn.p, 1);
outputVotes = ones(esn.d, 1) / esn.d;
errorIntegrated = zeros(esn.d,1);
netOut = ones(esn.p, esn.d)/ esn.p; 
netOut_originalScaling = netOut; 
in = ones(esn.p,1) / esn.p; 
votingPlotList = zeros(esn.d, initN + periodN * repeatN);
errPL = zeros(esn.d, initN + periodN * repeatN);
predictConfidencePL = zeros(esn.d, initN + periodN * repeatN);
netOutPL = zeros(esn.d, initN + periodN * repeatN);
generatedSignalPL = zeros(esn.p, initN + periodN * repeatN);
generatedSignalRawPL = zeros(1, initN + periodN * repeatN);
effectiveInPL = zeros(esn.p, initN + periodN * repeatN);
netOutCompletePL = zeros(esn.p, esn.d, initN + periodN * repeatN);
intWeights = esn.W * esn.spectralRadius; % adjust weights to desired SR

for iInput = 1:initN + periodN * repeatN    
    
    externalIn = triggerInput(iInput,:)'; %  is column vector
    in =  esn.inputScaling .* externalIn + esn.inputShift;
    
    if iInput <= initN + periodN * startAfterRepeatsN
        % this voting scheme simply means that only the external input is
        % used
        voting = [1; zeros(esn.d,1)];
    else
        % now we completely disregard external input and use the
        % combination of network outputs assembled via their vote weights
        % instead
        voting = [0; outputVotes];
    end
    votedData = [in netOut] * voting;      
    
    % restore perfect pitch coding
    votedData_originalScaling =  (votedData - esn.inputShift) ./ esn.inputScaling; 
    votedData_originalScaling = votedData_originalScaling / sum(votedData_originalScaling);
    generatedSignalPL(:, iInput) = votedData_originalScaling;
    votedData = votedData_originalScaling .* esn.inputScaling + esn.inputShift;
    
    
    % write voted combination of signals into input components of network
    % state. Apply the input scaling and add bias before doing so.    
    votedData_effectiveInput = esn.inputWeightScaling .* votedData + esn.bias;
    votedData_effectiveInput = votedData_effectiveInput + ...
        noiseLevel * (rand(esn.p,1) - 0.5);
    
    totalstate(esn.reservoirSize+1:end,1) = votedData_effectiveInput;    
    
    internalState = feval(esn.reservoirActivationFunction, [intWeights, esn.W_in] * totalstate );
%     internalState = internalState + noiseLevel * (rand(esn.reservoirSize, 1) - 0.5);
    
    totalstate(1:esn.reservoirSize,1) = internalState;    
    
    netOut = zeros(esn.p, esn.d);
    for n = 1:esn.d
        netOut(:,n) = (esn.W_out(:,:,n) * totalstate);
    end
    netOut = feval(esn.dataActivationFunction, netOut); 
    oldNetOut_originalScaling = netOut_originalScaling;
    netOut_originalScaling = (netOut - esn.teacherShift) ./ esn.teacherScaling;
    
    % per delay we wish to plot the re-constituted raw signal
    for d = 1:esn.d
        netOut_d = netOut_originalScaling(:,d);
        netOutReconstituted = unitCodedToRaw(netOut_d');
        netOutPL(d,iInput) = netOutReconstituted; 
    end
    netOutCompletePL(:,:,iInput) = netOut_originalScaling;
    
    % voting and competing
    
    errors_per_delay = zeros(esn.d, 1);
    for d = 1:esn.d
        netOut_d = oldNetOut_originalScaling(:,d);
        mse = mean((votedData_originalScaling - netOut_d).^2);
        errors_per_delay(d, 1) = mse;
    end   
    errorIntegrated = tanh(errorIntegrated - errDecay * errorIntegrated + errGrowthConstant * errors_per_delay);
    errPL(:,iInput) = errorIntegrated;
    % currentPredictConfidence is vector of size outputLength, giving in its i-th
    % component the predictConfidence level for that output unit in its own
    % prediction. Is 1 if prediction error is zero, is 0 if prediction
    % error is greater or equal than zeroPredictConfidenceThreshold, with linear
    % interpolation in between
    currentPredictConfidence = 1 - (1 / zeroPredictConfidenceThreshold) * errorIntegrated;
    currentPredictConfidence = max(0, currentPredictConfidence);
    predictConfidencePL(:,iInput) = currentPredictConfidence;
    
    % incrementing the voting 
    outputVotes = outputVotes - votingDecay * outputVotes ...
        + votingGrowthConstant * currentPredictConfidence ;
    outputVotes = outputVotes / sum(outputVotes);  
    votingPlotList(:,iInput) = outputVotes;
end

% diagnostics and plotting

generatedSignalRawPL = (unitCodedToRaw(generatedSignalPL'))';

figure(6); clf;
for i = 1:esn.d
    subplot(esn.d,1,i);
    plot(votingPlotList(i,:), 'LineWidth',2);
    for j = 0:startAfterRepeatsN
        line([initN + 1 + j * periodN, initN + 1 + j * periodN],[0; 1],'Color','k');
    end
    if i < esn.d
        set(gca, 'XTick',[], 'YTick',[], 'YLim',[0 1]);
    else
        set(gca, 'YTick',[0 1], 'YLim',[0 1]);
    end
    if i == 1
        title('voting states (shortest delay is top graph)');
    end
end

figure(7); clf;
for i = 1:esn.d
    subplot(esn.d,1,i);
    plot(errPL(i,:), 'LineWidth',2);
    for j = 0:startAfterRepeatsN
        line([initN + 1 + j * periodN, initN + 1 + j * periodN],[0; 1],'Color','k');
    end
    if i < esn.d
        set(gca, 'XTick',[], 'YTick',[]);
    else
        set(gca, 'YTick',[0 1]);
    end
    if i == 1
        title('tanh integrated err (shortest delay is top graph)');
    end
end


figure(8); clf;
for i = 1:esn.d
    subplot(esn.d,1,i);
    plot(predictConfidencePL(i,:), 'LineWidth',2);
    
    if i < esn.d
        set(gca, 'XTick',[], 'YTick',[]);
    else
        set(gca, 'YTick',[0 1]);
    end
    for j = 0:startAfterRepeatsN
        line([initN + 1 + j * periodN, initN + 1 + j * periodN],[0; 1],'Color','k');
    end
    if i == 1
        title('current predictConfidence (shortest delay is top graph)');
    end
end

figure(9); clf;
for i = 1:esn.d
    subplot(esn.d,1,i);
    plot(netOutPL(i,:));
    for j = 0:startAfterRepeatsN
        line([initN + 1 + j * periodN, initN + 1 + j * periodN],[0; 1],'Color','k');
    end
    if i == 1
        title('net Out (shortest delay is top graph)');
    end
end

figure(10); clf;
for p = 1:esn.p
    subplot(esn.p, 1, p);
    plot(generatedSignalPL(p,:)); hold on;
    for i = 0:startAfterRepeatsN
        line([initN + 1 + i * periodN, initN + 1 + i * periodN],[0; 1],'Color','k');
    end
    plot(triggerInput(:,p), 'r:'); hold off;
    if p == 1
        title('generatedSignal(b) trigger(r:)');
    end
end


figure(11); clf;
plot(generatedSignalRawPL, 'LineWidth',2);  hold on;
plot(triggerInputRaw,'r','LineWidth',1);  
for i = 0:startAfterRepeatsN
    line([initN + 1 + i * periodN, initN + 1 + i * periodN],[0; 1],'Color','k');
end
hold off;
set(gca, 'PlotBoxAspectRatio', [2 1 1]);
title('generated signal (blue thick)  trigger (red thin) ');


